/* --------------------------------------------------------------------------
$Id: write_vhlhdf.c,v 1.6 2000/03/22 08:43:06 ahenja Exp $
Program:        

Description:    Functions for writing an HDF5 file with the HighLevel
		protocol.

Author(s):      Anders Henja

Copyright:	Swedish Meteorological and Hydrological Institute, 2000

History:
$Log: write_vhlhdf.c,v $
Revision 1.6  2000/03/22 08:43:06  ahenja
Updated with spell checking and header information.

-----------------------------------------------------------------------------*/
#include "write_vhlhdf.h"

int doWriteHdf5Attribute(hid_t rootGrp, HL_Node* parentNode, char* parentName,
			 HL_Node* childNode, char* childName)
{
   hid_t tmpLocId=-1;

   if(!parentName || !childName) {
      if(_debug_hdf)
	 fprintf(stderr,"Can't write HDF5 group since either parentName or childName is NULL\n");
      return 0;
   }

   if(strcmp(parentName,"")==0) {
      tmpLocId=rootGrp;
   } else {
      tmpLocId=parentNode->hdfId;
   }

   if(strcmp(childNode->format,DATAFORMAT_UNDEFINED)!=0 && childNode->typeId<0) {
      childNode->typeId = translateCharToDatatype(childNode->format);
   }
   if(childNode->typeId<0) {
      if(_debug_hdf)
	 fprintf(stderr,"Can't recognize datatype '%s' or typeId='%ld'\n",
		 childNode->format,(long)childNode->typeId);
      return 0;
   }

   if(childNode->ndims==0) {
      if(writeScalarDataAttribute(tmpLocId, childNode->typeId, childName, childNode->data)<0) {
	 if(_debug_hdf)
	    fprintf(stderr,"Failed to write scalar data attribute '%s'\n",childNode->name);
	 return 0;
      }
   } else {
      if(writeSimpleDataAttribute(tmpLocId,childNode->typeId, childName,
				  childNode->ndims, childNode->dims,
				  childNode->data)<0) {
	 if(_debug_hdf)
	    fprintf(stderr,"Failed to write simple data attribute '%s'\n",childNode->name);
	 return 0;
      }
   }

   return 1;
}

int doWriteHdf5Group(hid_t rootGrp, HL_Node* parentNode, char* parentName,
		     HL_Node* childNode, char* childName)
{
   if(!parentName || !childName) {
      if(_debug_hdf)
	 fprintf(stderr,"Can't write HDF5 group since either parentName or childName is NULL\n");
      return 0;
   }
   if(strcmp(parentName,"")==0) {
      childNode->hdfId=H5Gcreate(rootGrp,childName,0);
   } else {
      childNode->hdfId=H5Gcreate(parentNode->hdfId,childName,0);
   }
   if(childNode->hdfId<0) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to create group %s\n",childNode->name);
      return 0;
   }
   return 1;
}

int doWriteHdf5Dataset(hid_t rootGrp, HL_Node* parentNode, char* parentName,
		       HL_Node* childNode, char* childName, int doCompress)
{
   hid_t tmpLocId=-1;

   if(!parentName || !childName) {
      if(_debug_hdf)
	 fprintf(stderr,"Can't write HDF5 dataset since either parentName or childName is NULL\n");
      return 0;
   }
   if(strcmp(parentName,"")==0) {
      tmpLocId=rootGrp;
   } else {
      tmpLocId=parentNode->hdfId;
   }

   if(strcmp(childNode->format,DATAFORMAT_UNDEFINED)!=0 && childNode->typeId<0) {
      childNode->typeId = translateCharToDatatype(childNode->format);
   }
   if(childNode->typeId<0) {
      if(_debug_hdf)
	 fprintf(stderr,"Can't recognize datatype '%s' or typeId='%ld'\n",
		 childNode->format,(long)childNode->typeId);
      return 0;
   }
   childNode->hdfId = createSimpleDataset(tmpLocId, childNode->typeId,
					  childName, childNode->ndims,
					  childNode->dims,
					  childNode->data, doCompress);
   if(childNode->hdfId<0) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to create dataset %s\n",childNode->name);
      return 0;
   }

   return 1;   
}

int doCommitHdf5Datatype(hid_t loc_id, HL_Node* parentNode, char* parentName,
			 HL_Node* childNode, char* childName)
{
   if(loc_id<0) {
      if(_debug_hdf)
	 fprintf(stderr,"Trying to create a committed datatype without setting typeid\n");
      return 0;
   }

   if((commitType(loc_id,childNode->name,childNode->hdfId))<0)
      return 0;
   if(H5Tcommitted(childNode->hdfId)<=0) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to commit datatype'%s'\n",childNode->name);
   }

   return 1;
}

int writeNodeList(HL_NodeList* nodelist, int doCompress)
{
   int i;
   HL_Node* parentNode=NULL;
   char parentName[256];
   char childName[256];
   hid_t file_id=-1;
   hid_t gid=-1;
   hid_t strhid=-1;

   if((file_id=createHlHdfFile(nodelist->filename))<0) {
      return 0;
   }
   if((gid=H5Gopen(file_id,"."))<0) {
      H5Fclose(file_id);
      return 0;
   }

   for(i=0;i<nodelist->nNodes;i++) {
      if(!extractParentChildName(nodelist->nodes[i],parentName,childName)) {
	 if(_debug_hdf)
	    fprintf(stderr,"Failed to extract parent, child name\n");
	 goto fail;
      }
      if((strcmp(parentName,"")!=0) && !(parentNode = getNode(nodelist,parentName))) {
	 if(_debug_hdf)
	    fprintf(stderr,"Failed to locate parent node '%s'\n",parentName);
	 goto fail;
      }

      switch(nodelist->nodes[i]->type) {
      case ATTRIBUTE_ID: {
	 if(!doWriteHdf5Attribute(gid,parentNode,parentName,nodelist->nodes[i],childName)) {
	    goto fail;
	 }
	 break;
      }
      case GROUP_ID: {
	 if(!doWriteHdf5Group(gid,parentNode, parentName, nodelist->nodes[i], childName)) { 
	    goto fail;
	 }
	 break;
      }
      case DATASET_ID: {
	 if(!doWriteHdf5Dataset(gid,parentNode, parentName, nodelist->nodes[i],childName,doCompress)) {
	    goto fail;
	 }
	 break;
      }
      case TYPE_ID: {
	 if(!doCommitHdf5Datatype(file_id,parentNode, parentName, nodelist->nodes[i],childName))
	    goto fail;
	 break;
      }
      default: {
	 fprintf(stderr,"Unrecognized type\n");
	 break;
      }
      }
   }

   if(gid>=0)
      H5Gclose(gid);
   if(file_id>=0)
      H5Fclose(file_id);
   if(strhid>=0)
      H5Tclose(strhid);
   return 1;
fail:
   if(gid>=0)
      H5Gclose(gid);
   if(file_id>=0)
      H5Fclose(file_id);
   if(strhid>=0)
      H5Tclose(strhid);
   return 0;
}
