/* --------------------------------------------------------------------------
$Id: pyhl.c,v 1.4 2000/08/03 11:26:57 dmichels Exp $
Program:        

Description:    The pyhl module source code, provides a python interface
		to the HighLevel protocol for writing and reading
		HDF5 files.

Author(s):      Anders Henja

Copyright:	Swedish Meteorological and Hydrological Institute, 2000

History:
$Log: pyhl.c,v $
Revision 1.4  2000/08/03 11:26:57  dmichels
Forgot to declare nbytes.

Revision 1.3  2000/08/02 12:10:41  dmichels
Removed use of PyArray_FromDimsAndData

Revision 1.2  2000/03/22 08:59:06  ahenja
Updated with spell checking and header information.
	
-----------------------------------------------------------------------------*/
#include <Python.h>
#include <arrayobject.h>

#include "pyhlhdf_common.h"
#include "read_vhlhdf.h"
#include "write_vhlhdf.h"

static PyObject *ErrorObject;

/* Some defines for error handling */
#define raiseException(type,msg)\
{PyErr_SetString(type,msg);return NULL;}

#define setException(type,msg)\
{PyErr_SetString(type,msg);}

typedef struct {
   PyObject_HEAD /*Always have to be on top*/
   HL_NodeList* nodelist;
} PyhlObject;

typedef struct {
   PyObject_HEAD /*Always have to be on top*/
   HL_Node* node;
} PyhlNode;

/*
 *Define that the PyhlObject is of the type Pyhl_Type
 */
staticforward PyTypeObject Pyhl_Type;
staticforward PyTypeObject PyhlNode_Type;

#define Pyhl_Check(op) ((op)->ob_type == &Pyhl_Type)
#define PyhlNode_Check(op) ((op)->ob_type == &PyhlNode_Type)

/* Pyhl module functions */
static void _dealloc(PyhlObject* pyhl_object)
{
   /*Nothing yet*/
   if(!pyhl_object)
      return;

   freeHL_NodeList(pyhl_object->nodelist);

   PyMem_DEL(pyhl_object);
}

static void _dealloc_pyhlnode(PyhlNode* pyhl_node)
{
   /*Nothing yet*/
   if(!pyhl_node)
      return;

   freeHL_Node(pyhl_node->node);

   PyMem_DEL(pyhl_node);
}

static PyObject* _pyhl_new_nodelist(PyObject* self, PyObject* args)
{
   PyhlObject* retv=NULL;
   retv=PyObject_NEW(PyhlObject,&Pyhl_Type);
   if(!retv)
      return NULL;

   if(!(retv->nodelist=newHL_NodeList())) {
      setException(PyExc_MemoryError,"Failed to create HL NodeList\n");
      _dealloc(retv);
      retv=NULL;
   }

   return (PyObject*)retv;
}

static PyObject* _pyhl_new_node(PyObject* self, PyObject* args)
{
   PyhlNode* retv=NULL;
   HL_Type type;
   char* nodename;

   if(!PyArg_ParseTuple(args,"is",&type,&nodename))
      return NULL;

   switch(type) {
   case ATTRIBUTE_ID:
   case GROUP_ID:
   case DATASET_ID:
   case TYPE_ID:
      break;
   default:
      printf("Unsupported node type %d\n",type);
      return NULL;
   }

   retv=PyObject_NEW(PyhlNode,&PyhlNode_Type);
   if(!retv)
      return NULL;

   switch(type) {
   case ATTRIBUTE_ID:
      retv->node = newHL_Attribute(nodename);
      break;
   case GROUP_ID:
      retv->node = newHL_Group(nodename);
      break;
   case DATASET_ID:
      retv->node = newHL_Dataset(nodename);
      break;
   case TYPE_ID:
      retv->node = newHL_Datatype(nodename);
      break;
   default:
      retv->node = NULL;
      break;
   }

   if(!(retv->node)) {
      setException(PyExc_MemoryError,"Failed to create Node\n");
      _dealloc_pyhlnode(retv);
      retv=NULL;
   }

   return (PyObject*)retv;
}

static PyObject* _pyhl_read_nodelist(PyObject* self, PyObject* args)
{
   HL_NodeList* nodelist=NULL;
   PyhlObject* retv=NULL;
   char* filename=NULL;
   char* frompath=NULL;
   
   if(!PyArg_ParseTuple(args,"s|s",&filename,&frompath))
      return NULL;

   if(!frompath) {
      nodelist = readHL_NodeList(filename);
   } else {
      nodelist = readHL_NodeListFrom(filename,frompath);
   }
   if(!nodelist) {
      char errmsg[256];
      if(!frompath)
	 sprintf("Could not read file '%s'",filename);
      else
	 sprintf("Could not read file '%s' from '%s'",filename,frompath);
      setException(PyExc_IOError,errmsg);
      goto fail;
   }

   if(!(retv=(PyhlObject*)_pyhl_new_nodelist(NULL,NULL))) {
      setException(PyExc_MemoryError,"Could not allocate nodelist instance");
      goto fail;
   }

   /*Change the nodelist*/
   freeHL_NodeList(retv->nodelist);
   retv->nodelist=nodelist;
   return (PyObject*)retv;
 fail:
   freeHL_NodeList(nodelist);
   if(retv)
      _dealloc(retv);
   return NULL;
}

/* PyhlObject member methods */
static PyObject* _pyhl_add_node(PyhlObject* self, PyObject* args)
{
   PyObject* inp;
   PyhlNode* pyhlNode;
   HL_Node* aNode;

   if(!PyArg_ParseTuple(args,"O",&inp))
      return NULL;

   if(!PyhlNode_Check(inp)) {
      setException(PyExc_TypeError,"Trying to add node, which not is of PyhlNodeCore type");
      return NULL;
   }

   pyhlNode=(PyhlNode*)inp;
   aNode=pyhlNode->node;

   if(!addNode(self->nodelist,aNode)) {
      setException(PyExc_MemoryError,"Could not add node to nodelist");
      freeHL_Node(aNode);
      return NULL;
   }

   pyhlNode->node=NULL;

   Py_INCREF(Py_None);
   return Py_None;
}

static PyObject* _pyhl_write(PyhlObject* self, PyObject* args)
{
   char* filename;
   int doCompress=6;
   if(!PyArg_ParseTuple(args,"s|i",&filename,&doCompress))
      return NULL;
   strcpy(self->nodelist->filename,filename);
   if(!writeNodeList(self->nodelist,doCompress)) {
      setException(PyExc_IOError,"Could not write file");
      return NULL;
   }
   Py_INCREF(Py_None);
   return Py_None;
}

static PyObject* _pyhl_get_node_names(PyhlObject* self, PyObject* args)
{
   PyObject* retv=NULL;
   PyObject* pyo=NULL;
   int i;
   char errbuf[256];
   if(!(retv=PyDict_New())) {
      setException(PyExc_MemoryError,"Could not allocate dictionary");
      return NULL;
   }
   for(i=0;i<self->nodelist->nNodes;i++) {
      switch(self->nodelist->nodes[i]->type) {
      case ATTRIBUTE_ID:
	 pyo=PyInt_FromLong(ATTRIBUTE_ID);
	 break;
      case DATASET_ID:
	 pyo=PyInt_FromLong(DATASET_ID);
	 break;
      case TYPE_ID:
	 pyo=PyInt_FromLong(TYPE_ID);
	 break;
      case GROUP_ID:
	 pyo=PyInt_FromLong(GROUP_ID);
	 break;
      default:
	 sprintf(errbuf,"Unknown type for node '%s': '%d'",self->nodelist->nodes[i]->name,
		 self->nodelist->nodes[i]->type);
	 setException(PyExc_TypeError,errbuf);
	 goto fail;
      }
      if(PyDict_SetItemString(retv,self->nodelist->nodes[i]->name,pyo)==-1) {
	 setException(PyExc_AttributeError,"Failed to set dictionary item");
	 goto fail;
      }
      Py_XDECREF(pyo);
      pyo=NULL;
   }

   return retv;
 fail:
   if(pyo)
      Py_XDECREF(pyo);
   if(retv)
      Py_XDECREF(retv);
   return NULL;
}

static PyObject* _pyhl_select_all(PyhlObject* self, PyObject* args)
{
   selectAllNodes(self->nodelist);
   Py_INCREF(Py_None);
   return Py_None;
}

static PyObject* _pyhl_select_node(PyhlObject* self, PyObject* args)
{
   char* nodename;
   char errbuf[256];

   if(!PyArg_ParseTuple(args,"s",&nodename))
      return NULL;
   if(!selectNode(self->nodelist,nodename)) {
      sprintf(errbuf,"Could not select node '%s'",nodename);
      setException(PyExc_AttributeError,errbuf);
      goto fail;
   }
   Py_INCREF(Py_None);
   return Py_None;
 fail:
   return NULL;
}

static PyObject* _pyhl_fetch(PyhlObject* self, PyObject* args)
{
   if(!fetchMarkedNodes(self->nodelist)) {
      setException(PyExc_IOError,"Could not fetch selected nodes");
      goto fail;
   }
   Py_INCREF(Py_None);
   return Py_None;
 fail:
   return NULL;
}

static PyObject* _pyhl_get_node(PyhlObject* self, PyObject* args)
{
   char* nodename;
   char errbuf[256];
   HL_Node* node=NULL;
   PyhlNode* retv=NULL;
   PyObject* myArgs=NULL;
   int i;
   size_t npts;

   if(!PyArg_ParseTuple(args,"s",&nodename))
      return NULL;

   if(!(node=getNode(self->nodelist,nodename))) {
      sprintf(errbuf,"Could not get node '%s'",nodename);
      setException(PyExc_IOError,errbuf);
      goto fail;
   }

   if(!(myArgs=Py_BuildValue("(is)",node->type,node->name))) {
      setException(PyExc_AttributeError,"Could not create argument tuple for allocating node");
      goto fail;
   }

   if(!(retv=(PyhlNode*)_pyhl_new_node((PyObject*)self,myArgs))) {
      setException(PyExc_AttributeError,"Could not create node to return");
      goto fail;
   }

   freeHL_Node(retv->node);
   retv->node=copyHL_Node(node);

   Py_XDECREF(myArgs);
   return (PyObject*)retv;
 fail:
   Py_XDECREF(myArgs);
   if(retv)
      _dealloc_pyhlnode(retv);
   return NULL;
}

/* PyhlNode member methods */
static PyObject* _pyhl_node_set_scalar_value(PyhlNode* self, PyObject* args)
{
   int itemSize;
   char* typename;
   PyObject* data;
   hid_t lhid=-1;
   hid_t tmpHid=-1;
   int chkVal=0;

   if(!PyArg_ParseTuple(args,"iOsl",&itemSize,&data,&typename,&lhid))
      return NULL;

   if(!self->node) {
      setException(PyExc_AttributeError,"The responsibility of the node has been dropped, probably by doing a addNode");
      return NULL;
   }

   if(strcmp(typename,"string")!=0 &&
      strcmp(typename,"compound")!=0 &&
      (tmpHid=translateCharToDatatype(typename))==-1) {
      char errmsg[256];
      sprintf(errmsg,"No datatype called '%s'.",typename);
      setException(PyExc_AttributeError,errmsg);
      return NULL;
   }

   if(tmpHid!=-1) {
      /* Well, ordinary attribute, just make sure the space is enough */
      size_t typeSize=H5Tget_size(tmpHid);
      chkVal=0;
      switch(H5Tget_class(tmpHid)) {
      case H5T_INTEGER: {
	 if(typeSize<=sizeof(char)) {
	    char mData = (char)PyInt_AsLong(data);
	    chkVal=setScalarValue(self->node,sizeof(mData),(unsigned char*)&mData,typename,-1);
	 } else if(typeSize<=sizeof(short)) {
	    short mData = (char)PyInt_AsLong(data);
	    chkVal=setScalarValue(self->node,sizeof(mData),(unsigned char*)&mData,typename,-1);
	 } else if(typeSize<=sizeof(int)) {
	    int mData = (int)PyInt_AsLong(data);
	    chkVal=setScalarValue(self->node,sizeof(mData),(unsigned char*)&mData,typename,-1);
	 } else if(typeSize<=sizeof(long)) {
	    long mData = (long)PyInt_AsLong(data);
	    chkVal=setScalarValue(self->node,sizeof(mData),(unsigned char*)&mData,typename,-1);
	 } else {
	    long long mData = (long long)PyLong_AsLongLong(data);
	    if(PyErr_Occurred()) {
	       setException(PyExc_AttributeError,"Could not translate to long long, probably not supported");
	       goto fail;
	    }
	    chkVal=setScalarValue(self->node,sizeof(mData),(unsigned char*)&mData,typename,-1);
	 }
	 break;
      }
      case H5T_FLOAT: {
	 if(typeSize<=sizeof(float)) {
	    float mData = (float)PyFloat_AsDouble(data);
	    chkVal=setScalarValue(self->node,sizeof(mData),(unsigned char*)&mData,typename,-1);
	 } else if (typeSize<=sizeof(double)) {
	    double mData = (double)PyFloat_AsDouble(data);
	    chkVal=setScalarValue(self->node,sizeof(mData),(unsigned char*)&mData,typename,-1);
	 } else {
	    fprintf(stderr,"Whoaa, greater float than double not supported\n");
	 }
	 break;
      }
      }
      H5Tclose(tmpHid);
      tmpHid=-1;
      if(!chkVal) {
	 setException(PyExc_TypeError,"Could not set scalar value");
	 goto fail;
      }
   } else if(strcmp(typename,"string")==0) {
      char* mData = PyString_AsString(data);
      if(!mData || !setScalarValue(self->node,strlen(mData)+1,(unsigned char*)mData,typename,-1)) {
	 setException(PyExc_ValueError,"Could not set scalar string");
	 goto fail;
      }
   } else if(strcmp(typename,"compound")==0) {
      char* mData = PyString_AsString(data);
      if(!mData || !setScalarValue(self->node,itemSize,(unsigned char*)mData,typename,lhid)) {
	 setException(PyExc_ValueError,"Could not set scalar compound value");
	 goto fail;
      }
   }

   if(tmpHid>=0)
      H5Tclose(tmpHid);
   Py_INCREF(Py_None);
   return Py_None;
 fail:
   if(tmpHid>=0)
      H5Tclose(tmpHid);

   return NULL;
}

static PyObject* _pyhl_node_set_array_value(PyhlNode* self, PyObject* args)
{
   PyObject* pydims;
   PyObject* data;
   hid_t lhid=-1;
   char* typename;
   int itemSize;
   hsize_t dims[4];
   int i;
   int ndim;
   PyObject* pyo;
   size_t tmpSize;

   if(!PyArg_ParseTuple(args,"iOOsl",&itemSize,&pydims,&data,&typename,&lhid))
      return NULL;

   if(!self->node) {
      setException(PyExc_AttributeError,"The responsibility of the node has been dropped, probably by doing a addNode");
      return NULL;
   }

   if(PyObject_Length(pydims)>4) {
      setException(PyExc_ValueError,"Not more than 4 dimensions are allowed");
      goto fail;
   }

   if(PyArray_Check(data) && isFormatSupported(typename)) {
      /*Ahh plain array*/
      if(PyObject_Length(pydims)!=((PyArrayObject*)data)->nd) {
	 setException(PyExc_ValueError,"Rank of Array != no of dims");
	 goto fail;
      }
      ndim=PyObject_Length(pydims);
      for(i=0;i<ndim;i++) { /*Extract dimensions*/
	 if(!(pyo=PySequence_GetItem(pydims,i))) {
	    setException(PyExc_AttributeError,"Could not get list item");
	    goto fail;
	 }
	 dims[i] = PyInt_AsLong(pyo);
	 Py_XDECREF(pyo);
	 pyo=NULL;
	 if(dims[i]!=((PyArrayObject*)data)->dimensions[i]) {
	    setException(PyExc_AttributeError,"Array dimensions != list dimensions");
	    goto fail;
	 }
      }
      if((tmpSize = whatSizeIsHdfFormat(typename))<0) {
	 setException(PyExc_ValueError,"Could not determine size");
	 goto fail;
      }
      if(tmpSize!=((PyArrayObject*)data)->descr->elsize) {
	 setException(PyExc_ValueError,"Type sizes are different between format and array");
	 goto fail;
      }
      if(!setArrayValue(self->node,tmpSize,ndim,dims,(unsigned char*)((PyArrayObject*)data)->data,
			typename,-1)) {
	 setException(PyExc_AttributeError,"Could not set array data");
	 goto fail;
      }
   } else if(strcmp(typename,"compound")==0) {
      size_t tmpLen=0;
      char* tmpData=PyString_AsString(data);
      if(!tmpData) {
	 setException(PyExc_AttributeError,"Could not translate data to a string");
	 goto fail;
      }
      ndim=PyObject_Length(pydims);
      for(i=0;i<ndim;i++) {
	 if(!(pyo=PySequence_GetItem(pydims,i))) {
	    setException(PyExc_AttributeError,"Could not get list item");
	    goto fail;
	 }
	 dims[i] = PyInt_AsLong(pyo);
	 Py_XDECREF(pyo);
	 pyo=NULL;
      }
      tmpLen=1;
      for(i=0;i<ndim;i++) {
	 tmpLen*=dims[i];
      }
      tmpLen*=itemSize;
      if(PyObject_Length(data)!=tmpLen) {
	 setException(PyExc_AttributeError,"There is some inconsistency between data size and calculated datasize");
	 goto fail;
      }
      if(!setArrayValue(self->node,itemSize,ndim,dims,(unsigned char*)tmpData,typename,lhid)) {
	 setException(PyExc_AttributeError,"Could not set array data");
	 goto fail;
      }
   } else {
      setException(PyExc_AttributeError,"Only support arrays and compound types right now");
      goto fail;
   }

   if(pyo) Py_XDECREF(pyo);
   Py_INCREF(Py_None);
   return Py_None;
 fail:
   if(pyo) Py_XDECREF(pyo);
   return NULL;
}

static PyObject* _pyhl_node_commit(PyhlNode* self, PyObject* args)
{
   hid_t dType;
   if(!PyArg_ParseTuple(args,"l",&dType))
      return NULL;
   if(!self->node->type==TYPE_ID) {
      setException(PyExc_AttributeError,"Trying to commit a node which not is a type node");
      return NULL;
   }
   commitDatatype(self->node,dType);
   Py_INCREF(Py_None);
   return Py_None;
}

static PyObject* _pyhl_node_name(PyhlNode* self, PyObject* args)
{
   return PyString_FromString(self->node->name);
}

static PyObject* _pyhl_node_dims(PyhlNode* self,PyObject* args)
{
   PyObject* retv=NULL;
   int i;
   PyObject* pyo=NULL;
   if(!(retv=PyList_New(0))) {
      return NULL;
   }
   for(i=0;i<self->node->ndims;i++) {
      if(!(pyo = PyInt_FromLong(self->node->dims[i]))) {
	 setException(PyExc_ValueError,"Could not create py integer");
	 goto fail;
      }
      if(PyList_Append(retv,pyo)==-1) {
	 setException(PyExc_ValueError,"Could not append list item");
	 goto fail;
      }
      Py_XDECREF(pyo);
      pyo=NULL;
   }
   Py_XDECREF(pyo);
   return retv;
 fail:
   Py_XDECREF(pyo);
   Py_XDECREF(retv);
   return NULL;
}

static PyObject* _pyhl_node_format(PyhlNode* self,PyObject* args)
{
   return PyString_FromString(self->node->format);
}

static PyObject* _pyhl_node_data(PyhlNode* self, PyObject* args)
{
   PyObject* retv=NULL;
   hid_t tmpHid=-1;
   size_t typeSize;
   char errbuf[256];
   int* dims=NULL;
   int i;
   size_t npts;

   if(self->node->typeId>=0) {
      tmpHid=H5Tcopy(self->node->typeId);
   } else if(isFormatSupported(self->node->format)) {
      tmpHid=translateCharToDatatype(self->node->format);
   } else {
      setException(PyExc_AttributeError,"Strange type, can't handle");
      return NULL;
   }

   typeSize = H5Tget_size(tmpHid);

   if(self->node->ndims==0) { /*Scalar*/
      switch(H5Tget_class(tmpHid)) {
      case H5T_INTEGER: {
	 if(typeSize<=sizeof(char)) {
	    char v;
	    memcpy(&v,self->node->data,typeSize);
	    retv=PyInt_FromLong((long)v);
	 } else if(typeSize<=sizeof(short)) {
	    short v;
	    memcpy(&v,self->node->data,typeSize);
	    retv=PyInt_FromLong((long)v);
	 } else if(typeSize<=sizeof(int)) {
	    int v;
	    memcpy(&v,self->node->data,typeSize);
	    retv=PyInt_FromLong((long)v);
	 } else if(typeSize<=sizeof(long)) {
	    long v;
	    memcpy(&v,self->node->data,typeSize);
	    retv=PyInt_FromLong((long)v);
	 } else {
	    long long v;
	    memcpy(&v,self->node->data,typeSize);
	    retv=PyLong_FromLongLong(v);
	 }
	 break;
      }
      case H5T_FLOAT: {
	 if(typeSize<=sizeof(float)) {
	    float v;
	    memcpy(&v,self->node->data,typeSize);
	    retv=PyFloat_FromDouble((double)v);
	 } else if(typeSize<=sizeof(double)) {
	    double v;
	    memcpy(&v,self->node->data,typeSize);
	    retv=PyFloat_FromDouble(v);
	 } else {
	    fprintf(stderr,"Whoaa, greater float than double not supported\n");
	 }
	 break;
      }
      case H5T_COMPOUND: {
	 retv=PyString_FromStringAndSize((char*)self->node->data,typeSize);
	 break;
      }
      case H5T_STRING: {
	 retv=PyString_FromStringAndSize((char*)self->node->data,typeSize);
	 break;
      }
      default: {
	 sprintf(errbuf,"Unrecognized class %d",(int)H5Tget_class(tmpHid));
	 setException(PyExc_TypeError,errbuf);
	 break;
      }
      }
   } else { /*Simple*/
      if(!(dims = malloc(sizeof(int)*self->node->ndims))) {
	 setException(PyExc_MemoryError,"Could not allocate dims");
	 goto fail;
      }
      switch(H5Tget_class(tmpHid)) {
      case H5T_INTEGER:
      case H5T_FLOAT: {
	 int iformat=pyarraytypeFromHdfType(self->node->format);
	 int nbytes;
	 if(iformat==-1) {
	    sprintf(errbuf,"Unrecognized datatype %s",self->node->format);
	    setException(PyExc_TypeError,errbuf);
	    goto fail;
	 }
	 for(i=0;i<self->node->ndims;i++)
	    dims[i]=(int)self->node->dims[i];

	 /* Replaced PyArray_FromDimsAndData with this */
	 retv = PyArray_FromDims(self->node->ndims,dims,iformat);
	 if(!retv) {
	    setException(PyExc_MemoryError,"Could not create array");
	    goto fail;
	 }
	 nbytes = 1;
	 for(i=0;i<self->node->ndims;i++)
	    nbytes *= dims[i];
	 nbytes *= ((PyArrayObject*)retv)->descr->elsize;

	 memcpy(((PyArrayObject*)retv)->data,(unsigned char*)self->node->data,nbytes);

	 /* Just read that this is unsecure */
	 /*retv = PyArray_FromDimsAndData(self->node->ndims,dims,iformat,(char*)self->node->data);*/
	 break;
      }
      case H5T_COMPOUND:
      case H5T_STRING: {
	 npts=1;
	 for(i=0;i<self->node->ndims;i++)
	    npts*=self->node->dims[i];
	 npts*=typeSize;
	 retv=PyString_FromStringAndSize((char*)self->node->data,npts);
	 break;
      }
      default: {
	 sprintf(errbuf,"Unrecognized class %d",(int)H5Tget_class(tmpHid));
	 break;
      }
      }
   }
   if(dims)
      free(dims);
   return retv;
 fail:
   Py_XDECREF(retv);
   if(dims)
      free(dims);
   return NULL;
}

static struct PyMethodDef methods[] =
{
  {"addNode",(PyCFunction)_pyhl_add_node,1},
  {"write",(PyCFunction)_pyhl_write,1},
  {"getNodeNames",(PyCFunction)_pyhl_get_node_names,1},
  {"selectAll",(PyCFunction)_pyhl_select_all,1},
  {"selectNode",(PyCFunction)_pyhl_select_node,1},
  {"fetch",(PyCFunction)_pyhl_fetch,1},
  {"getNode",(PyCFunction)_pyhl_get_node,1},
  {NULL, NULL} /* sentinel */
};

static struct PyMethodDef node_methods[] =
{
   {"setScalarValue",(PyCFunction)_pyhl_node_set_scalar_value,1},
   {"setArrayValue",(PyCFunction)_pyhl_node_set_array_value,1},
   {"commit",(PyCFunction)_pyhl_node_commit,1},

   /* Inquiry options */
   {"name",(PyCFunction)_pyhl_node_name,1},
   {"dims",(PyCFunction)_pyhl_node_dims,1},
   {"format",(PyCFunction)_pyhl_node_format,1},
   {"data",(PyCFunction)_pyhl_node_data,1},
   {NULL, NULL} /* sentinel */
};

static PyObject* _getattr(PyhlObject* self, char* name)
{
    PyObject* res;

    if(strcmp(name,"ATTRIBUTE_ID")==0)
       return PyInt_FromLong(ATTRIBUTE_ID);
    if(strcmp(name,"GROUP_ID")==0)
       return PyInt_FromLong(GROUP_ID);
    if(strcmp(name,"DATASET_ID")==0)
       return PyInt_FromLong(DATASET_ID);
    if(strcmp(name,"TYPE_ID")==0)
       return PyInt_FromLong(TYPE_ID);

    res = Py_FindMethod(methods, (PyObject*) self, name);
    if (res)
	return res;

    PyErr_Clear();

    PyErr_SetString(PyExc_AttributeError, name);

    return NULL;
}

static PyObject* _getattr_node(PyhlNode* self, char* name)
{
    PyObject* res;

    res = Py_FindMethod(node_methods, (PyObject*) self, name);
    if (res)
	return res;

    PyErr_Clear();

    PyErr_SetString(PyExc_AttributeError, name);

    return NULL;
}

statichere PyTypeObject Pyhl_Type = {
    PyObject_HEAD_INIT(NULL)
    0,				/*ob_size*/
    "PyhlCore",		/*tp_name*/
    sizeof(PyhlObject),	/*tp_size*/
    0,				/*tp_itemsize*/
    /* methods */
    (destructor)_dealloc,	/*tp_dealloc*/
    0,				/*tp_print*/
    (getattrfunc)_getattr,	/*tp_getattr*/
    0,				/*tp_setattr*/
    0,				/*tp_compare*/
    0,				/*tp_repr*/
    0,                          /*tp_as_number */
    0,
    0,                          /*tp_as_mapping */
    0                           /*tp_hash*/
};

statichere PyTypeObject PyhlNode_Type = {
    PyObject_HEAD_INIT(NULL)
    0,				/*ob_size*/
    "PyhlNodeCore",		/*tp_name*/
    sizeof(PyhlNode),	/*tp_size*/
    0,				/*tp_itemsize*/
    /* methods */
    (destructor)_dealloc_pyhlnode,/*tp_dealloc*/
    0,				/*tp_print*/
    (getattrfunc)_getattr_node,	/*tp_getattr*/
    0,				/*tp_setattr*/
    0,				/*tp_compare*/
    0,				/*tp_repr*/
    0,                          /*tp_as_number */
    0,
    0,                          /*tp_as_mapping */
    0                           /*tp_hash*/
};

static PyMethodDef functions[] = {
  {"nodelist", (PyCFunction)_pyhl_new_nodelist, 1},
  {"node", (PyCFunction)_pyhl_new_node, 1},
  {"read_nodelist",(PyCFunction)_pyhl_read_nodelist,1},
  {NULL,NULL} /*Sentinel*/
};

void init_pyhl()
{
   PyObject *m,*d;
   PyObject *tmp;
   Pyhl_Type.ob_type=&PyType_Type;
   PyhlNode_Type.ob_type=&PyType_Type;

   m=Py_InitModule("_pyhl",functions);
   d=PyModule_GetDict(m);

   tmp = PyInt_FromLong(ATTRIBUTE_ID);
   PyDict_SetItemString(d,"ATTRIBUTE_ID",tmp);
   Py_XDECREF(tmp);

   tmp = PyInt_FromLong(GROUP_ID);
   PyDict_SetItemString(d,"GROUP_ID",tmp);
   Py_XDECREF(tmp);

   tmp = PyInt_FromLong(DATASET_ID);
   PyDict_SetItemString(d,"DATASET_ID",tmp);
   Py_XDECREF(tmp);

   tmp = PyInt_FromLong(TYPE_ID);
   PyDict_SetItemString(d,"TYPE_ID",tmp);
   Py_XDECREF(tmp);

   import_array(); /*To make sure I get access to Numeric*/
   /*Always have to do this*/
   initHlHdf();
   /*And this I just do to be able to get debugging info from hdf*/
   debugHlHdf(2);
}
